/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    Asynch_IO_Impl.h
 *
 *  Asynch_IO_Impl.h,v 4.19 2002/08/27 17:35:09 shuston Exp
 *
 *
 *  This class contains asbtract base classes for all the concrete
 *  implementation classes for the various asynchronous operations
 *  that are used with the Praoctor.
 *
 *
 *  @author Irfan Pyarali (irfan@cs.wustl.edu)
 *  @author Tim Harrison (harrison@cs.wustl.edu)
 *  @author Alexander Babu Arulanthu <alex@cs.wustl.edu>
 *  @author Roger Tragin <r.tragin@computer.org>
 *  @author Alexander Libman <alibman@ihug.com.au>
 */
//=============================================================================

#ifndef TPROACTOR_ASYNCH_IO_IMPL_H 
#define TPROACTOR_ASYNCH_IO_IMPL_H 

#include /**/ "ace/pre.h"

#include "TProactor/TPROACTOR_Export.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) || (defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION))
// This only works on Win32 platforms and on Unix platforms supporting
// aio calls.

#include "TProactor/Asynch_IO.h"

// Forward declaration.
class TRB_Proactor_Impl;

/**
 * @class TRB_Asynch_Result_Impl
 *
 * @brief Abstract base class for the all the classes that provide
 * concrete implementations for TRB_Asynch_Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Result_Impl
{
public:

  /// Result State
  enum State 
    {
      INITIALIZED = 0,  // emtpy object
      STARTED     = 1,  // operation in progress
      DEFERRED    = 2,  // start deferred,i.e unable to start,
                        // but it is possible to start later
      START_ERROR = 3,  // unable to start (no point to make DEFERRED)
      RESULT_READY= 4,  // operation completed,i.e.
                        // result (transfer count,error and 
                        // completion is key) is stored in member variables,
                        // but not dispatched yet 
      POSTED      = 5,  // operation was posted via post_completion.
                        // we assume  result is ready 
      COMPLETED   = 6   // operation fully completed,i.e. result is ready
                        // and it is time to dispatch
    };

  /// Operation code
  enum Op
    {
      NO_OP         = 0x0000,
      READ_STREAM   = 0x0001,
      WRITE_STREAM  = 0x0002,
      READ_FILE     = 0x0004,   
      WRITE_FILE    = 0x0008,  
      READ_DGRAM    = 0x0010,  
      WRITE_DGRAM   = 0x0020, 
      ACCEPT        = 0x0040,      
      CONNECT       = 0x0080,
      TRANSMIT_FILE = 0x0100,
      TIMER         = 0x0200,
      USER_OP       = 0x8000
    };     

  /// Special internal flags
  enum 
    {
      NORMAL_RESULT   = 0, // Normal result

      DUMMY_RESULT    = 1, // Dummy result, can be used as head node in 
                           // ACE_Double_Linked_List<>

      INTERNAL_RESULT = 2  // Internal AIO result.
                           // Should be never deleted (WIN32/POSIX)
                           // and never passed
                           // to the Completion Dispatcher (POSIX).
                           // Dispatched internally.
                           // Useful for Notifications :
                           // Pipe_Interrupt_Strategy,
                           // Proactor Message queues, etc.
      
    };

  virtual ~TRB_Asynch_Result_Impl (void);

  /// The following 5 methods now is not virtual!
  /**
   * This returns the ACT associated with the handle when it was
   * registered with the I/O completion port.  This ACT is not the
   * same as the ACT associated with the asynchronous operation.
   */
  const void *completion_key (void) const;

  /// ACT associated with the operation.
  const void *act (void) const;

  /// Did the operation succeed?
  int success (void) const;

  /// Number of bytes transferred by the operation.
  size_t bytes_transferred (void) const;
  
  /// Error value if the operation fail.
  u_long error (void) const;

  /// Deprecated, should be used set_ready(...) now
  /// Simulate error value to use in the post_completion ()
  void set_error (u_long errcode);

  /// Deprecated, should be used set_ready(...) now
  /// Simulate value to use in the post_completion ()
  void set_bytes_transferred (size_t nbytes);


  /// Get state
  int state (void) const;

  /// Get Operation code
  int op_code (void) const;

  /// Get Flags
  int is_internal_result (void) const;

  /// Set state STARTED
  int set_started ();

  /// Set state START_ERROR
  int set_start_error (u_long errcode);

  /// Set state DEFERRED (currently used in POSIX)
  int set_deferred ();

  /// Set state RESULT_READY
  int set_ready (ssize_t nbytes,
                 u_long errcode,
                 const void *completion_key);

  /// Set state POSTED 
  int set_posted ();

  /// Set state COMPLETED
  int set_completed ();
  
  /// Internal Result can be reused, i.e. restarted again
  int reuse ();

  int is_read_family (void) const;
  int is_write_family (void) const;


  /// Event associated with the OVERLAPPED structure.
  virtual ACE_HANDLE event (void) const = 0;

  /// This really make sense only when doing file I/O.
  virtual u_long offset (void) const = 0;
  virtual u_long offset_high (void) const = 0;

  /// Priority of the operation.
  virtual int priority (void) const = 0;

  /**
   * POSIX4 real-time signal number to be used for the
   * operation. <signal_number> ranges from SIGRTMIN to SIGRTMAX. By
   * default, SIGRTMIN is used to issue <aio_> calls. This is a no-op
   * on non-POSIX4 systems and returns 0.
   */
  virtual int signal_number (void) const = 0;

  // protected:
  //
  // These two should really be protected.  But sometimes it
  // simplifies code to be able to "fake" a result.  Use carefully.
  /// This is called when the asynchronous operation completes.
  virtual void complete (size_t bytes_transferred,
                         int success,
                         const void *completion_key,
                         u_long error = 0) = 0;

  /// Post <this> to the Proactor's completion port.
  virtual int post_completion (TRB_Proactor_Impl *proactor) = 0;

  /// for debug
  void dump();

protected:
  /// Constructor.
  TRB_Asynch_Result_Impl ();
 

  /// ACT for this operation.
  const void *act_;

  /// ACT associated with handle.
  const void *completion_key_;

  /// Bytes transferred by this operation.
  size_t bytes_transferred_;

  /// Error if operation failed.
  u_long error_;

  /// Success indicator, replaced by State
  /// int success_;
  State state_;

  /// Operation code
  Op    op_code_;

  /// special flags 
  int flags_;
};

/**
 * @class TRB_Asynch_Operation_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Operation.
 */
class TPROACTOR_Export TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Operation_Impl (void);

  /**
   * Initializes the factory with information which will be used with
   * each asynchronous call.  If (<handle> == ACE_INVALID_HANDLE),
   * <TRB_Handler::handle> will be called on the <handler> to get the
   * correct handle.
   */
  virtual int open (TRB_Handler &handler,
                    ACE_HANDLE handle,
                    const void *completion_key,
                    TRB_Proactor *proactor) = 0;


  virtual int close (void) = 0;

  /**
   * This cancels all pending accepts operations that were issued by
   * the calling thread.  The function does not cancel asynchronous
   * operations issued by other threads.
   */
  virtual int cancel (void) = 0;

  // = Access methods.

  /// Return the underlying proactor.
  virtual TRB_Proactor* proactor (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Operation_Impl (void);
};

/**
 * @class TRB_Asynch_Read_Stream_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Read_Stream
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_Stream_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Read_Stream_Impl (void);

  /// This starts off an asynchronous read. Upto <bytes_to_read> will
  /// be read and stored in the <message_block>.
  virtual int read (ACE_Message_Block &message_block,
                    size_t bytes_to_read,
                    const void *act,
                    int priority,
                    int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with scatter support, through chaining of composite
  * message blocks using the continuation field.
  */
  virtual int readv (ACE_Message_Block &message_block,
                     size_t bytes_to_read,
                     const void *act,
                     int priority,
                     int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_Stream_Impl (void);
};

/**
 * @class TRB_Asynch_Read_Stream_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Read_Stream::Result class.
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_Stream_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Read_Stream_Result_Impl (void);

  /// The number of bytes which were requested at the start of the
  /// asynchronous read.
  virtual size_t bytes_to_read (void) const = 0;

  /// Message block which contains the read data.
  virtual ACE_Message_Block &message_block (void) const = 0;

  /// I/O handle used for reading.
  virtual ACE_HANDLE handle (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_Stream_Result_Impl ();
};

/**
 * @class TRB_Asynch_Write_Stream_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Write_Stream class.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_Stream_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Write_Stream_Impl (void);

  /// This starts off an asynchronous write.  Upto <bytes_to_write>
  /// will be written from the <message_block>.
  virtual int write (ACE_Message_Block &message_block,
                     size_t bytes_to_write,
                     const void *act,
                     int priority,
                     int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with gather support, through chaining of composite
  * message blocks using the continuation field.
  */
  virtual int writev (ACE_Message_Block &message_block,
                      size_t bytes_to_write,
                      const void *act,
                      int priority,
                      int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */
  
protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_Stream_Impl (void);
};

/**
 * @class TRB_Asynch_Write_Stream_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Write_Stream::Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_Stream_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Write_Stream_Result_Impl (void);

  /// The number of bytes which were requested at the start of the
  /// asynchronous write.
  virtual size_t bytes_to_write (void) const = 0;

  /// Message block that contains the data to be written.
  virtual ACE_Message_Block &message_block (void) const = 0;

  /// I/O handle used for writing.
  virtual ACE_HANDLE handle (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_Stream_Result_Impl ();
};

/**
 * @class TRB_Asynch_Read_File_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Read_File::Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_File_Impl : public virtual TRB_Asynch_Read_Stream_Impl
{
public:
  virtual ~TRB_Asynch_Read_File_Impl (void);

  /**
   * This starts off an asynchronous read.  Upto <bytes_to_read> will
   * be read and stored in the <message_block>.  The read will start
   * at <offset> from the beginning of the file.
   */
  virtual int read (ACE_Message_Block &message_block,
                    size_t bytes_to_read,
                    u_long offset,
                    u_long offset_high,
                    const void *act,
                    int priority,
                    int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with scatter support, through chaining of composite
  * message blocks using the continuation field.
  * NOTE: In win32 Each data block payload must be at least the size of a system 
  * memory page and must be aligned on a system memory page size boundary
  */
  virtual int readv (ACE_Message_Block &message_block,
                     size_t bytes_to_read,
                     u_long offset,
                     u_long offset_high,
                     const void *act,
                     int priority,
                     int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

  // We don;t need to redefine the following function again because it
  // has already been defined in TRB_Asynch_Read_Stream_Impl.  But we
  // still need it here to supress a overwriting pure virtual function
  // warning in KAI compiler.
  /// This starts off an asynchronous read. Upto <bytes_to_read> will
  /// be read and stored in the <message_block>.
  virtual int read (ACE_Message_Block &message_block,
                    size_t bytes_to_read,
                    const void *act,
                    int priority,
                    int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with scatter support, through chaining of composite
  * message blocks using the continuation field.
  */
  virtual int readv (ACE_Message_Block &message_block,
                     size_t bytes_to_read,
                     const void *act,
                     int priority,
                     int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_File_Impl (void);
};

/**
 * @class TRB_Asynch_Read_File_Result_Impl
 *
 * @brief This is the abstract base class for all the concrete
 * implementation classes for TRB_Asynch_Read_File::Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_File_Result_Impl : public virtual TRB_Asynch_Read_Stream_Result_Impl
{
public:
  /// Destructor.
  virtual ~TRB_Asynch_Read_File_Result_Impl (void);

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_File_Result_Impl ();
};

/**
 * @class TRB_Asynch_Write_File_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Write_File.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_File_Impl : public virtual TRB_Asynch_Write_Stream_Impl
{
public:
  virtual ~TRB_Asynch_Write_File_Impl (void);

  /**
   * This starts off an asynchronous write.  Upto <bytes_to_write>
   * will be write and stored in the <message_block>.  The write will
   * start at <offset> from the beginning of the file.
   */
  virtual int write (ACE_Message_Block &message_block,
                     size_t bytes_to_write,
                     u_long offset,
                     u_long offset_high,
                     const void *act,
                     int priority,
                     int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with gather support, through chaining of composite
  * message blocks using the continuation field.
  * NOTE: In win32 Each data block payload must be at least the size of a system 
  * memory page and must be aligned on a system memory page size boundary
  */
  virtual int writev (ACE_Message_Block &message_block,
                      size_t bytes_to_write,
                      u_long offset,
                      u_long offset_high,
                      const void *act,
                      int priority,
                      int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */


  // We don;t need to redefine the following function again because it
  // has already been defined in TRB_Asynch_Write_Stream_Impl.  But we
  // still need it here to supress a overwriting pure virtual function
  // warning in KAI compiler.
  /// This starts off an asynchronous write.  Upto <bytes_to_write>
  /// will be written from the <message_block>.
  virtual int write (ACE_Message_Block &message_block,
                     size_t bytes_to_write,
                     const void *act,
                     int priority,
                     int signal_number) = 0;

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
  /**
  * Same as above but with gather support, through chaining of composite
  * message blocks using the continuation field.
  */
  virtual int writev (ACE_Message_Block &message_block,
                      size_t bytes_to_write,
                      const void *act,
                      int priority,
                      int signal_number) = 0;
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_File_Impl (void);
};

/**
 * @class TRB_Asynch_Write_File_Result_Impl
 *
 * @brief This is the abstract base class for all the concrete
 * implementation classes that provide different implementations
 * for the TRB_Asynch_Write_File::Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_File_Result_Impl : public virtual TRB_Asynch_Write_Stream_Result_Impl
{
public:
  virtual ~TRB_Asynch_Write_File_Result_Impl (void);

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_File_Result_Impl ();
};

/**
 * @class TRB_Asynch_Accept_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Accept.
 *
 */
class TPROACTOR_Export TRB_Asynch_Accept_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Accept_Impl (void);

  /**
   * This starts off an asynchronous accept.  The asynchronous accept
   * call also allows any initial data to be returned to the
   * <handler>.  Upto <bytes_to_read> will be read and stored in the
   * <message_block>.  The <accept_handle> will be used for the
   * <accept> call.  If (<accept_handle> == INVALID_HANDLE), a new
   * handle will be created.
   *
   * <message_block> must be specified. This is because the address of
   * the new connection is placed at the end of this buffer.
   */
  virtual int accept (ACE_Message_Block &message_block,
                      size_t bytes_to_read,
                      ACE_HANDLE accept_handle,
                      const void *act,
                      int priority,
                    int signal_number) = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Accept_Impl (void);
};

/**
 * @class TRB_Asynch_Accept_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Accept.
 *
 */
class TPROACTOR_Export TRB_Asynch_Accept_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Accept_Result_Impl (void);

  /// The number of bytes which were requested at the start of the
  /// asynchronous accept.
  virtual size_t bytes_to_read (void) const = 0;

  /// Message block which contains the read data.
  virtual ACE_Message_Block &message_block (void) const = 0;

  /// I/O handle used for accepting new connections.
  virtual ACE_HANDLE listen_handle (void) const = 0;

  /// I/O handle for the new connection.
  virtual ACE_HANDLE accept_handle (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Accept_Result_Impl ();
};


/**
 * @class TRB_Asynch_Connect_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Connect.
 *
 */
class TPROACTOR_Export TRB_Asynch_Connect_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Connect_Impl (void);

  /**
   * This starts off an asynchronous connect
   */
  virtual int connect (ACE_HANDLE connect_handle,
                       const ACE_Addr & remote_sap,
                       const ACE_Addr & local_sap,
                       int   reuse_addr,
                       const void *act,
                       int   priority,
                       int   signal_number) = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Connect_Impl (void);
};

/**
 * @class TRB_Asynch_Connect_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Connect.
 *
 */
class TPROACTOR_Export TRB_Asynch_Connect_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Connect_Result_Impl (void);

  /// I/O handle for the connection.
  virtual ACE_HANDLE connect_handle (void) const = 0;

  /// The address to connect
  virtual const ACE_Addr& remote_address () const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Connect_Result_Impl ();
};


/**
 * @class TRB_Asynch_Transmit_File_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Transmit_File.
 *
 */
class TRB_Asynch_Transmit_File_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Transmit_File_Impl (void);

  /// This starts off an asynchronous transmit file.
  virtual int transmit_file (ACE_HANDLE file,
                             TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer,
                             size_t bytes_to_write,
                             u_long offset,
                             u_long offset_high,
                             size_t bytes_per_send,
                             u_long flags,
                             const void *act,
                             int priority,
                             int signal_number) = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Transmit_File_Impl (void);
};

/**
 * @class TRB_Asynch_Transmit_File_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Transmit_File::Result.
 *
 */
class TPROACTOR_Export TRB_Asynch_Transmit_File_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Transmit_File_Result_Impl (void);

  /// Socket used for transmitting the file.
  virtual ACE_HANDLE socket (void) const = 0;

  /// File from which the data is read.
  virtual ACE_HANDLE file (void) const = 0;

  /// Header and trailer data associated with this transmit file.
  virtual TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer (void) const = 0;

  /// The number of bytes which were requested at the start of the
  /// asynchronous transmit file.
  virtual size_t bytes_to_write (void) const = 0;

  /// Number of bytes per send requested at the start of the transmit
  /// file.
  virtual size_t bytes_per_send (void) const = 0;

  /// Flags which were passed into transmit file.
  virtual u_long flags (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Transmit_File_Result_Impl ();
};


/**
 * @class TRB_Asynch_Read_Dgram_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Read_Dgram
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_Dgram_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Read_Dgram_Impl (void);

  /** This starts off an asynchronous read.  Upto
   * <message_block->total_size()> will be read and stored in the
   * <message_block>.  <message_block>'s <wr_ptr> will be updated to reflect
   * the added bytes if the read operation is successful completed.
   * Return code of 1 means immediate success and <number_of_bytes_recvd>
   * will contain number of bytes read.  The <TRB_Handler::handle_read_dgram>
   * method will still be called.  Return code of 0 means the IO will
   * complete proactively.  Return code of -1 means there was an error, use
   * errno to get the error code.
   *
   * Scatter/gather is supported on WIN32 by using the <message_block->cont()>
   * method.  Up to ACE_IOV_MAX <message_block>'s are supported.  Upto 
   * <message_block->size()> bytes will be read into each <message block> for
   * a total of <message_block->total_size()> bytes.  All <message_block>'s
   * <wr_ptr>'s will be updated to reflect the added bytes for each 
   * <message_block>
   *
   * Priority of the operation is specified by <priority>. On POSIX4-Unix,
   * this is supported. Works like <nice> in Unix. Negative values are not
   * allowed. 0 means priority of the operation same as the process
   * priority. 1 means priority of the operation is one less than
   * process. And so forth. On Win32, <priority> is a no-op.
   * <signal_number> is the POSIX4 real-time signal number to be used
   * for the operation. <signal_number> ranges from ACE_SIGRTMIN to
   * ACE_SIGRTMAX. This argument is a no-op on non-POSIX4 systems.
   */
  virtual ssize_t recv (ACE_Message_Block *message_block,
                        size_t &number_of_bytes_recvd,
                        int flags,
                        int protocol_family,
                        const void *act,
                        int priority,
                        int signal_number) = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_Dgram_Impl (void);
};

/**
 * @class TRB_Asynch_Read_Dgram_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Read_Dgram::Result class.
 *
 */
class TPROACTOR_Export TRB_Asynch_Read_Dgram_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Read_Dgram_Result_Impl (void);

  /// Message block which contains the read data
  virtual ACE_Message_Block *message_block (void) const = 0;

  /// The number of bytes which were requested at the start of the
  /// asynchronous read.
  virtual size_t bytes_to_read (void) const = 0;

  /// The address of where the packet came from
  virtual const ACE_Addr &  remote_address () const = 0;
  virtual int remote_address (ACE_Addr& addr) const = 0;

  /// The flags used in the read
  virtual int flags (void) const = 0;

  /// I/O handle used for reading.
  virtual ACE_HANDLE handle (void) const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Read_Dgram_Result_Impl ();
};

/**
 * @class TRB_Asynch_Write_Dgram_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Write_Dgram class.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_Dgram_Impl : public virtual TRB_Asynch_Operation_Impl
{
public:
  virtual ~TRB_Asynch_Write_Dgram_Impl (void);

  /** This starts off an asynchronous send.  Upto
   * <message_block->total_length()> will be sent.  <message_block>'s 
   * <rd_ptr> will be updated to reflect the sent bytes if the send operation
   * is successful completed.
   * Return code of 1 means immediate success and <number_of_bytes_sent>
   * is updated to number of bytes sent.  The <TRB_Handler::handle_write_dgram>
   * method will still be called.  Return code of 0 means the IO will
   * complete proactively.  Return code of -1 means there was an error, use
   * errno to get the error code.
   *
   * Scatter/gather is supported on WIN32 by using the <message_block->cont()>
   * method.  Up to ACE_IOV_MAX <message_block>'s are supported.  Upto 
   * <message_block->length()> bytes will be sent from each <message block>
   * for a total of <message_block->total_length()> bytes.  All
   * <message_block>'s <rd_ptr>'s will be updated to reflect the bytes sent
   * from each <message_block>.
   *
   * Priority of the operation is specified by <priority>. On POSIX4-Unix,
   * this is supported. Works like <nice> in Unix. Negative values are not
   * allowed. 0 means priority of the operation same as the process
   * priority. 1 means priority of the operation is one less than
   * process. And so forth. On Win32, this argument is a no-op.
   * <signal_number> is the POSIX4 real-time signal number to be used
   * for the operation. <signal_number> ranges from ACE_SIGRTMIN to
   * ACE_SIGRTMAX. This argument is a no-op on non-POSIX4 systems.
   */
  virtual ssize_t send (ACE_Message_Block *message_block,
                        size_t &number_of_bytes_sent,
                        int flags,
                        const ACE_Addr &addr,
                        const void *act,
                        int priority,
                        int signal_number) = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_Dgram_Impl (void);
};

/**
 * @class TRB_Asynch_Write_Dgram_Result_Impl
 *
 * @brief Abstract base class for all the concrete implementation
 * classes that provide different implementations for the
 * TRB_Asynch_Write_Dgram::Result class.
 *
 */
class TPROACTOR_Export TRB_Asynch_Write_Dgram_Result_Impl : public virtual TRB_Asynch_Result_Impl
{
public:
  virtual ~TRB_Asynch_Write_Dgram_Result_Impl (void);

  /// The number of bytes which were requested at the start of the
  /// asynchronous write.
  virtual size_t bytes_to_write (void) const = 0;

  /// Message block which contains the sent data
  virtual ACE_Message_Block *message_block (void) const = 0;

  /// The flags using in the write
  virtual int flags (void) const = 0;

  /// I/O handle used for writing.
  virtual ACE_HANDLE handle (void) const = 0;

  /// The address to send
  virtual const ACE_Addr &  remote_address () const = 0;

protected:
  /// Do-nothing constructor.
  TRB_Asynch_Write_Dgram_Result_Impl ();
};

#if defined (__ACE_INLINE__)
#include "TProactor/Asynch_IO_Impl.inl"
#endif /* __ACE_INLINE__ */

#endif /* ACE_HAS_AIO_CALLS  || !ACE_HAS_WINCE && ACE_WIN32 */
#include "ace/post.h"
#endif //TPROACTOR_ASYNCH_IO_IMPL_H 
