/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id: $

#include "TProactor/POSIX_AIO_Dedicated_Processor.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "TProactor/POSIX_AIO_Provider.h"
#include "TProactor/POSIX_AIO_Dispatcher.h"

#include "ace/Log_Msg.h"

typedef ACE_Reverse_Lock<ACE_SYNCH_MUTEX> REVERSE_MUTEX;

TRB_POSIX_AIO_Dedicated_Processor::TRB_POSIX_AIO_Dedicated_Processor
  (TRB_POSIX_AIO_Dispatcher & dispatcher,
   const TRB_POSIX_AIO_Config & config)
  : TRB_POSIX_AIO_Processor (dispatcher, config),
    ACE_Task <ACE_SYNCH>(&dispatcher.thread_manager()),
    svc_state_ (SS_INIT),
    flg_stop_ (0),
    non_empty_condition_ (mutex_),
    ready_condition_ (mutex_)
{
}

TRB_POSIX_AIO_Dedicated_Processor::~TRB_POSIX_AIO_Dedicated_Processor (void)
{
  switch (this->svc_state_)
    {
    case SS_INIT:
    case SS_STOPPED:
      break;
    default:
      ACE_ASSERT (0);
    }
}

/*
 *******************************************************************
 * svc() 
 * should be called only for dedicated leader model
 *
 *
 *******************************************************************
 */
int
TRB_POSIX_AIO_Dedicated_Processor::svc (void)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, 
                            ace_mon,
                            this->mutex_,
                            -1));


  ACE_DEBUG((LM_DEBUG,
            ACE_LIB_TEXT("(%t) Dedicated Processor Start\n")));

  if ((this->init_provider () == 0) &&
      (this->acquire_leadership () == 1))
    {
      this->ready_condition_.signal ();

      while (this->flg_stop_ == 0)
        {
          if (this->num_total_aios () == 0)
            {
              this->svc_state_ = SS_WAIT_COND_VAR;
              this->non_empty_condition_.wait ();
            }
          else
            {
              this->svc_state_ = SS_HANDLE_EVENTS;
              this->handle_aio_events (0); 
            }
        }
    }

  ACE_DEBUG((LM_DEBUG,
            ACE_LIB_TEXT("(%t) Dedicated Processor Stop\n")));

  this->svc_state_ = SS_STOPPING;
  
  this->release_leadership ();
  this->fini_provider ();

  //in case init_provider() failed
  this->ready_condition_.signal ();
  
  return 0;
}

/**
 ****************************************************************
 * virtual methods
 ****************************************************************
 */
int 
TRB_POSIX_AIO_Dedicated_Processor::start (void)
{
  switch (this->svc_state_)
    {
    case SS_INIT:
    case SS_STOPPED:
      break;
    default:
      return -1;
    }

  this->svc_state_ = SS_ACTIVATING;

  if (this->activate (THR_NEW_LWP | THR_JOINABLE, 1) != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT ("%N:%l:%p\n"),
                       ACE_LIB_TEXT ("TRB_POSIX_AIO_Dedicated_Processor::start failed 1")),
                       -1);

  if (this->ready_condition_.wait () != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT ("%N:%l:%p\n"),
                       ACE_LIB_TEXT ("TRB_POSIX_AIO_Dedicated_Processor::start failed 2")),
                       -1);
  return 0;
}

int 
TRB_POSIX_AIO_Dedicated_Processor::stop (void)
{
  switch (this->svc_state_)
    {
    case SS_INIT:
    case SS_STOPPED:
      return 0;
    default:
      break;
    }

  this->flg_stop_ = 1;
  this->activate_leader (IR_BREAK);

    {
      REVERSE_MUTEX anti_mutex (this->mutex_);
      ACE_MT (ACE_GUARD_RETURN (REVERSE_MUTEX, ace_mon, anti_mutex, -1));
     
      int rc = this->wait ();

      if  (rc != 0)
        ACE_ERROR_RETURN 
          ((LM_ERROR,
            ACE_LIB_TEXT ("%N:%l:%p\n"),
            ACE_LIB_TEXT ("TRB_POSIX_AIO_Dedicated_Processor::stop failed")),
            -1);
    }

  this->svc_state_ = SS_STOPPED;

  return 0;
}

int 
TRB_POSIX_AIO_Dedicated_Processor::is_dedicated_leadership (void) const
{
  return 1;
}  

int
TRB_POSIX_AIO_Dedicated_Processor::is_ready_for_leadership (void) const
{
  // This method is called when we about to be a leader.
  // Purpose - ask processor if it really needs leadership.
  // return : 1 - processor is ready to the leadership
  //          0 - processor is not ready (example, no active aios)

  return 1; // always ready
}

int
TRB_POSIX_AIO_Dedicated_Processor::activate_leader (int reason)
{
  ACE_ASSERT (!this->is_in_leader_context());

  switch (this->svc_state_)
    {
    case SS_WAIT_COND_VAR:
      if (this->non_empty_condition_.signal() !=0)
        ACE_ERROR_RETURN ((LM_ERROR, 
          ACE_LIB_TEXT("(%t) TRB_POSIX_AIO_Dedicated_Processor::")
          ACE_LIB_TEXT("interrupt_leader: cond. signal failed\n")),
          -1);

      return 0;
      
    case SS_HANDLE_EVENTS:
      return this->interrupt_leader (reason);

    default:
      break;
    }

  return 0;
}

#endif /* ACE_HAS_AIO_CALLS */
