/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    POSIX_AIO_Dispatcher.h
 *
 *  $Id: $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef TPROACTOR_POSIX_AIO_DISPATCHER_H 
#define TPROACTOR_POSIX_AIO_DISPATCHER_H 

#include "TProactor/TPROACTOR_Export.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/Synch_T.h"
#include "ace/Thread_Manager.h"
#include "TProactor/POSIX_AIO_Strategies.h"

//#include "ace/Handle_Set.h"

class Thread_Dispatcher;

typedef ACE_Double_Linked_List<Thread_Dispatcher> Thread_Dispatcher_List;
typedef ACE_Double_Linked_List_Iterator<Thread_Dispatcher> Thread_Dispatcher_Iterator;

class Thread_Dispatcher
{
private:
  // For dummy node 
  // ACE_Double_Linked_List
  static ACE_SYNCH_MUTEX dummy_mutex_;

  friend class TRB_POSIX_AIO_Dispatcher;
  friend class ACE_Double_Linked_List<Thread_Dispatcher>;
  friend class ACE_Double_Linked_List_Iterator_Base<Thread_Dispatcher>;
  friend class ACE_Double_Linked_List_Iterator<Thread_Dispatcher>;

  Thread_Dispatcher (TRB_POSIX_AIO_Dispatcher * dispatcher = 0);
  virtual ~Thread_Dispatcher ();


  void unlink (void);
  void link (Thread_Dispatcher_List & list);

  void push_result (TRB_POSIX_Asynch_Result * result);
  TRB_POSIX_Asynch_Result * pop_result(void );

  //
  TRB_POSIX_AIO_Dispatcher * dispatcher_;
  Thread_Dispatcher_List *   owner_;

  //
  ACE_SYNCH_CONDITION condition_;

  //
  TRB_POSIX_Asynch_Result * disp_result_;
  TRB_POSIX_Asynch_Waiting_List list_waiters_;

  TRB_POSIX_AIO_Processor * processor_;

  Thread_Dispatcher * next_;
  Thread_Dispatcher * prev_;

  unsigned int num_events_;
};


//Forward declaration
class TRB_POSIX_AIO_Processor;

/**
 * @class TRB_POSIX_AIO_Dispatcher
 *
 * @brief POSIX implementation .
 */
class TPROACTOR_Export TRB_POSIX_AIO_Dispatcher
{
  friend class TRB_POSIX_AIO_Processor;


public:


  enum Post_Notify_Flags 
    {
      NOTIFY_NONE     = 0, // post result only does queueing result
      NOTIFY_FOLLOWER = 1, // post result notifies only follower
      NOTIFY_STRONG   = 2  // post result can interrupt leader
                           // if there are no followers and
                           // there is an active leader
    };


  /// Clear the internal results queue
  /// Thread-unsafe operation, should be called only
  /// with locked proactor mutex_
  int clear_result_queue (void);

  /// Put "posted/ready" result into the internal queue
  /// and notify completion dispatcher.
  /// Parameter notify_flag specifies how to notify 
  /// the completion dispatcher.
  /// Thread-unsafe operation, should be called only
  /// with locked proactor mutex_
  /// Thread-safe version is 
  /// POSIX_Proactor::post_completion() method
  int post_result (TRB_POSIX_Asynch_Result *result, int notify_flg);

  /// Get "post_completed" result from the internal queue
  /// Thread-unsafe operation, should be called only
  /// with locked proactor mutex_
  TRB_POSIX_Asynch_Result * getq_result (void);

  /// Notify Completion Dispatcher
  /// If there are no followers, then interrupt shared leader
  /// (if exists) with purpose to release leadership, because
  /// we have completed results for dispatching
  int notify(int notify_flg);


  int handle_events (ACE_Time_Value * wait_time);

  int init (TRB_POSIX_AIO_Processor * prime_processor,
            TRB_POSIX_AIO_Processor * second_processor);
  
  int fini ();

  ACE_SYNCH_MUTEX & mutex ();  
  
  ACE_Thread_Manager & thread_manager();


  void dump (const ACE_TCHAR * txt = 0);

  void get_loading (size_t & num_followers,
                    size_t & num_dispatchers,
                    size_t & queue_size) const;

  /// Constructor.
  TRB_POSIX_AIO_Dispatcher (ACE_SYNCH_MUTEX & mutex);

  /// Virtual destructor.
  virtual ~TRB_POSIX_AIO_Dispatcher (void);

protected:

  /// Notify Completion Dispatcher
  /// If there are no followers, then do nothing.
  int notify_follower(void);

  int notify_leader(void);

  /// Process the internal results queue
  /// Thread-unsafe operation, should be called only
  /// with locked proactor mutex_
  int process_result_queue (Thread_Dispatcher & thr_dispatcher);

  /**
   * Protect against structured exceptions caused by user code when
   * dispatching handles. The <completion_key> is not very useful
   * compared to <AST> that can be associated each asynchronous
   * operation. <completion_key> is implemented right now for the
   * POSIX Proators.
   * The old name is application_specific_code()
   */
  int dispatch_result (TRB_POSIX_Asynch_Result * result);
  
  void promote_leader (TRB_POSIX_AIO_Processor * processor);


  /// Wait as follower
  int wait_as_follower (ACE_Time_Value * wait_time,
                        Thread_Dispatcher & thr_dispatcher);

  /// Leader method
  int handle_aio_events (ACE_Time_Value * wait_time,
                         TRB_POSIX_AIO_Processor * processor,
                         Thread_Dispatcher & thr_dispatcher);


  int is_handle_in_dispatch(ACE_HANDLE handle,
                            TRB_POSIX_AIO_Processor * processor);
 
  /// reference to proactor mutex
  ACE_SYNCH_MUTEX & mutex_;
  
  /// Thread manager for AIO_Dedicated_Processor's threads
  ACE_Thread_Manager thr_mngr_;

  // ACE_SYNCH_CONDITION condition_;
  // ACE_SYNCH_SEMAPHORE condition_;

  /// List of completed/post-completed AIOs
  TRB_POSIX_Asynch_Waiting_List completed_list_;
  
  size_t num_active_threads_;
  
  TRB_POSIX_AIO_Processor * prime_processor_;
  TRB_POSIX_AIO_Processor * second_processor_;

private:

   Thread_Dispatcher * allocate_thr_dispatcher (void);
   void free_thr_dispatcher (Thread_Dispatcher * thr_dispatcher);

   Thread_Dispatcher_List leaders_;
   Thread_Dispatcher_List dispatchers_;
   Thread_Dispatcher_List followers_;
   Thread_Dispatcher_List free_list_;


   Thread_Dispatcher * find_dispatcher(ACE_HANDLE handle);

};



#endif /* ACE_HAS_AIO_CALLS */

#endif //TPROACTOR_POSIX_AIO_DISPATCHER_H 


