/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_Poll_Strategy.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/Log_Msg.h"
#include "ace/Addr.h"
#include "ace/ACE.h"

/**
 *****************************************************************
 *  POSIX Poll based AIO Wait Strategy
 *
 *****************************************************************
 */
TRB_POSIX_AIO_Poll_Strategy::~TRB_POSIX_AIO_Poll_Strategy()
{
  this->fini ();
}

TRB_POSIX_AIO_Poll_Strategy::TRB_POSIX_AIO_Poll_Strategy 
  (TRB_POSIX_AIO_Select_Provider & provider)
  : TRB_POSIX_AIO_Select_Strategy_Base (provider),
    poll_fds_   (0),
    next_fds_   (0),
    num_events_ (0)
{
}

int 
TRB_POSIX_AIO_Poll_Strategy::init (void)
{
  if (this->alloc_repository (ACE_AIO_NO_LIMIT, 0) != 0)
    return -1;;

  
  ACE_NEW_RETURN (this->poll_fds_,
                  pollfd[this->size_],
                  -1);

  ACE_DEBUG ((LM_DEBUG,
    ACE_LIB_TEXT("(%t)TRB_POSIX_AIO_Poll_Strategy size=%d\n"), 
    (int) this->size_));


  return 0;
}  

int 
TRB_POSIX_AIO_Poll_Strategy::fini (void)
{
  delete [] this->poll_fds_;

  this->poll_fds_   = 0;
  this->next_fds_  = 0;
  this->num_events_ = 0;
  
  this->free_repository ();

  return 0;
}

int 
TRB_POSIX_AIO_Poll_Strategy::on_aio_start (TRB_POSIX_Asynch_Result * result)
{
  ACE_HANDLE handle = result->aio_fildes;

  int blk_mode = ACE_BIT_DISABLED (ACE::get_flags (handle), ACE_NONBLOCK);

  // Set the handle into non-blocking mode if it's not already in it.
  if (blk_mode)
    ACE::set_flags (handle, ACE_NONBLOCK);

  return 0; 
}


int
TRB_POSIX_AIO_Poll_Strategy::prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_TRACE ("TRB_POSIX_AIO_Poll_Strategy::prepare_to_wait");

  if (this->num_events_ > 0)
    return WAIT_COMPLETED; 


  size_t num_op = started_list.size();

  this->next_fds_ = this->poll_fds_;
  
  u_int i = 0;

  for ( i=0; i < this->size_ && this->num_events_ < num_op; ++i)
    {
      this->next_fds_->fd = ACE_INVALID_HANDLE;
      this->next_fds_->events = 0;
      this->next_fds_->revents = 0;
      
      if (this->pending_reads_[i] != 0 && 
          this->pending_reads_[i] != IO_READY)
        {
          this->next_fds_->fd = i;
          this->next_fds_->events |= POLLIN;
        }

      if (this->pending_writes_[i] != 0 && 
          this->pending_writes_[i] != IO_READY)
        {
          this->next_fds_->fd = i;
          this->next_fds_->events |= POLLOUT;
        }

      if (this->next_fds_->events != 0)
        {
          ++this->next_fds_;
          ++this->num_events_;
        }
    }

  this->next_fds_ = this->poll_fds_;


  return WAIT_ENABLED;
}

int
TRB_POSIX_AIO_Poll_Strategy::wait (ACE_Time_Value * wait_time)
{
  ACE_TRACE ("TRB_POSIX_AIO_Poll_Strategy::wait");

  long timeout = -1L;
  
  if (wait_time != 0 )
  {
     timeout =  ACE_static_cast (long, wait_time->msec ());
  }
  


  this->fd_info_.set (ACE_INVALID_HANDLE, 0);
     
  int rc = ::poll (this->poll_fds_, 
                   this->num_events_, 
                   ACE_static_cast(int, timeout));
 
  if (rc == 0)
    {
      this->num_events_ = 0;
      return WAIT_TIMEOUT;
    }

  if (rc  > 0)
    {
      this->fd_info_.set (this->next_fds_->fd,
                          this->next_fds_->revents);
      
      return WAIT_COMPLETED;
    }


   switch (errno)
    {
    case EINTR :    // Interrupted call
      this->num_events_ = 0;
      return WAIT_INTERRUPTED;

    default:
      break;
    }

  this->fd_info_.set (this->next_fds_->fd,
                      this->next_fds_->revents);
      
  return WAIT_ERROR;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_AIO_Poll_Strategy::get_next_completed_result 
  (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_UNUSED_ARG (started_list);

  for (;;)
    {

      ACE_HANDLE handle = this->fd_info_.handle();

      if (handle == ACE_INVALID_HANDLE)
        return 0;

      int rc = this->check (this->fd_info_);

      if (rc)
        {
          TRB_POSIX_Asynch_Result * result = this->fd_info_.get_result();
          ACE_ASSERT (result !=0);
          return result;
        }

      if ( --this->num_events_ <= 0 )
        break;


      ++this->next_fds_ ;
      
      this->fd_info_.set (this->next_fds_->fd,
                          this->next_fds_->revents);

    }

  return 0;
}

// *****************************************************************
//  Declare interest
//  Return code :
//    1 - we should interrupt leader to update wait state
//    0 - we do not have to interrupt leader
// *****************************************************************

int
TRB_POSIX_AIO_Poll_Strategy::declare_interest (ACE_HANDLE handle, 
                                                   u_int mask)
{
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (mask);

  return 1;
}

int
TRB_POSIX_AIO_Poll_Strategy::clear_interest (ACE_HANDLE handle, 
                                                 u_int mask)
{  
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (mask);

  return 0;
}


#endif /* ACE_HAS_AIO_CALLS */
