/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    POSIX_AIO_Strategies.h
 *
 *  $Id:  $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef TRB_POSIX_AIO_STRATEGIES_H 
#define TRB_POSIX_AIO_STRATEGIES_H 

#include "TProactor/POSIX_AIO_Config.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/Synch_T.h"
#include "TProactor/POSIX_Asynch_IO.h"
#include "ace/Message_Block.h"
#include "ace/Pipe.h"


//Forward declarations

class TRB_POSIX_AIO_Processor;
class TRB_POSIX_AIO_Provider;
class TRB_POSIX_AIO_Wait_Strategy;
class TRB_POSIX_AIO_Interrupt_Strategy;


/**
 * @class TRB_POSIX_AIO_Wait_Strategy
 *  
 */
class TRB_POSIX_AIO_Wait_Strategy  
{
public:

  friend class TRB_POSIX_AIO_Processor;
  friend class TRB_POSIX_AIO_Provider;


  enum Wait_Retcode
    {
      WAIT_TIMEOUT      = 0, // timeout or repeat leater ETIME,EAGAIN
      WAIT_ENABLED      = WAIT_TIMEOUT,
      WAIT_INTERRUPTED  = 1, // wait was interrupted (for example, EINTR)
      WAIT_COMPLETED    = 2, // Completion happened
      WAIT_ERROR        = -1 // Some errors
    };

  virtual ~TRB_POSIX_AIO_Wait_Strategy(void);

  TRB_POSIX_AIO_Wait_Strategy (TRB_POSIX_AIO_Provider & provider);

  TRB_POSIX_AIO_Provider & provider (void) const;
  const TRB_POSIX_AIO_Config & config (void) const;

protected:

  virtual  int init (void);
  virtual  int fini (void);

  /// factory method - create interrupt strategy
  /// Parameter is_type - desired interrupt strategy type.
  /// If Wait_Strategy use Pipe/Signal Interrupt_Strategies,
  /// it can use default implementation

  virtual TRB_POSIX_AIO_Interrupt_Strategy * create_interrupt_strategy (void);
  
  /// If Wait_Strategy use Pipe/Signal Interrupt_Strategies,
  /// it can use default implementation
  virtual 
    void destroy_interrupt_strategy (TRB_POSIX_AIO_Interrupt_Strategy * i_strategy);
 
  /// This methods is called when aio is about to start
  /// Wait_Stategy can update aiocb.aio_sigevent
  /// to setup OS notification mechanizm
  virtual int on_aio_start (TRB_POSIX_Asynch_Result * result) =0;


  /// This is the the last method which is called with
  /// locked Proactor mutex before calling wait ().
  /// It is safe to save neccessary information
  /// into member variables that shared
  /// between : prepare_to_wait(),wait(), after_wait().
  /// get_next_completed_result()
  /// AIO_Processor guarantees that other threads never
  /// will call any of these methods.
  /// Also, AIO_Processor guarantees that
  /// iterator will be unchanged between this calls
  /// Return  codes:
  ///   WAIT_COMPLETED   Wait strategy has some completed results already,
  ///                    don't need to call wait() 
  ///   WAIT_ENABLED     Strategy is ready to wait for completions 
  ///   WAIT_ERROR       Errors
  virtual int prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list)=0;

  /// Wait for AIO completion
  /// Attention!!!! this is only one method that is called
  /// with unlocked Proactor mutex
  /// Returns Wait_Retcode
   virtual int wait (ACE_Time_Value * wait_time) =0;

  /// Called after wait, Proactor mutex is locked again
  /// called while there are comleted results.
  /// Return code:
  ///  Pointer to the completed result
  ///  NULL - no more completed results
  virtual TRB_POSIX_Asynch_Result * 
    get_next_completed_result (TRB_POSIX_Asynch_Waiting_List & started_list)=0;

  TRB_POSIX_AIO_Provider & provider_;
};

/**
 * @class TRB_POSIX_AIO_Interrupt_Strategy
 *  This abstract class defines
 *  Interrupt Strategy interface used 
 *  to interrupt Wait_Strategy
 *
 */
class TPROACTOR_Export TRB_POSIX_AIO_Interrupt_Strategy
{
public:
  friend class TRB_POSIX_AIO_Processor;
  friend class TRB_POSIX_AIO_Provider;


  virtual ~TRB_POSIX_AIO_Interrupt_Strategy (void);
  TRB_POSIX_AIO_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider);

  TRB_POSIX_AIO_Provider & provider (void) const;
  const TRB_POSIX_AIO_Config & config (void) const;


protected:

  virtual  int init (void) =0;
  virtual  int fini (void) =0;

  /// This virtual function is called by
  /// AIO_Processor immediately after 
  /// the current thread obtained the leadership.
  /// To obtain leader/current thread id
  /// use ACE_Thread::self(void)
  
  virtual  int on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor) =0;

  /// This virtual function is called by
  /// AIO_Processor when the current thread is
  /// leader  is  about to leave leadership.
  /// To obtain leader/current thread id
  /// use ACE_Thread::self(void)
  virtual  int on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor) =0;

  /// This virtual function is called by AIO_Processor
  /// from non-leader thread when AIO_Processor
  /// needs to interrupt the leader thread.
  /// AIO_Processor never calls this method
  /// from the leader thread.
  virtual  int interrupt(ACE_thread_t leader_id) =0;

  TRB_POSIX_AIO_Provider & provider_;
};

/**
 * @class TRB_POSIX_AIO_Signal_Interrupt_Strategy
 * This is univeral interrupt_stategy suitable
 * to interrupt any wait_strategy, that can recognize
 * error EINTR during wait call.
 */
class TPROACTOR_Export TRB_POSIX_AIO_Signal_Interrupt_Strategy
  : public TRB_POSIX_AIO_Interrupt_Strategy
{
public:

  TRB_POSIX_AIO_Signal_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider);
  virtual ~TRB_POSIX_AIO_Signal_Interrupt_Strategy (void);

protected:

  virtual  int init (void);
  virtual  int fini (void);
  virtual  int on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor);
  virtual  int on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor);
  virtual  int interrupt(ACE_thread_t leader_id);


  static void sig_handler (int sig_num,
                           siginfo_t *,
                           ucontext_t *);

  int setup_signal_handler (int sig_num) const;

  // how one of:= SIG_BLOCK,SIG_UNBLOCK,SIG_SETMASK
  int mask_signals (int how, const sigset_t *signals) const;

  //signal number that is used as interrupt signal
  int signal_number_;

  //signal mask. used to unblock/block signal
  //when leader is on_leader_activatedd/deon_leader_activatedd
  sigset_t signal_mask_;
};



/**
 * @class TRB_POSIX_AIO_Pipe_Interrupt_Strategy
 * Formely Notify_Pipe_Manager.
 * This is universal strategy, that can be combined with 
 * any Wait_Strategy. It uses notification Pipe
 * Optimization: this strategy always reuses 
 * the same TRB_POSIX_Asynch_Result to avoid
 * extra new(void)/delete(void) operators. This result
 * should be marked as INTERNAL_AIO, so
 * AIO_Processor will never pass it to completion
 * dispatcher.
 */

class TPROACTOR_Export TRB_POSIX_AIO_Pipe_Interrupt_Strategy
  : public TRB_POSIX_AIO_Interrupt_Strategy,
    public TRB_Handler
{
public:

  TRB_POSIX_AIO_Pipe_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider);
  virtual ~TRB_POSIX_AIO_Pipe_Interrupt_Strategy (void);

protected:

  virtual  int init (void);
  virtual  int fini (void);
  virtual  int on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor);
  virtual  int on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor);
  virtual  int interrupt(ACE_thread_t leader_id);


  class TRB_POSIX_AIO_Interrupt_Result : public TRB_POSIX_Asynch_Read_Stream_Result
  {
    public:
    /// Constructor.
    TRB_POSIX_AIO_Interrupt_Result (TRB_POSIX_AIO_Pipe_Interrupt_Strategy & owner);
    
    /// Destructor.
    virtual ~TRB_POSIX_AIO_Interrupt_Result (void);

    /// virtual, overriden, defined in base class
    void complete (size_t bytes_transferred,
                   int success,
                   const void *completion_key,
                   u_long error);
  };

  friend class TRB_POSIX_AIO_Interrupt_Result;

  /// Notification pipe 
  ACE_Pipe pipe_;

  /// Message block to read from  the pipe.
  ACE_Message_Block message_block_;
  char              read_buffer_ [256];

  /// To do asynch_read on the pipe.
  TRB_POSIX_AIO_Interrupt_Result * notify_result_;

  u_long num_wr_;
  u_long bytes_wr_;
  u_long num_rd_;
  u_long bytes_rd_;
};

#endif /* ACE_HAS_AIO_CALLS */

#endif //TRB_POSIX_AIO_STRATEGIES_H 
